/* Copyright 2007-2011, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* Syscon memory map:
 * 0x00: reads 0x7700
 * 0x02: Control register:
     bits 3-0:   FPGA revision (RO)
     bits 5-4:   Mode2, Mode1
     bits 7-6:   Scratch register
     bit 8:      Enable CAN (override DIO 15,16)
     bit 9:      Enable 2nd CAN (override DIO 10,11)
     bit 10:     Enable SPI (override DIO 17-20)
     bit 11:     Enable 12.5MHz base board clock (override DIO 3)
     bit 12:     Enable UART0 TXEN (override DIO 12)
     bit 13:     Enable UART4 TXEN (override DIO 14)
     bit 14:     Enable touchscreen (override DIO 30-35)
     bit 15:     Reset switch enable
 * 0x04: external bus config register
 * 0x06: watchdog feed register
 * 0x08: free running 1MHz counter LSB
 * 0x0a: free running 1MHz counter MSB
 * 0x0c: RNG LSB
 * 0x0e: RNG MSB
 * 0x12: 
     bit 11: Green LED
     bit 12: Red LED
 * 0x16: 
     bit 12:     Enable UART3 TXEN (override DIO 13)
     bit 13:     Enable UART5 TXEN (override DIO 7)
     bit 14:     Enable UART1 TXEN (override DIO 8)
     bit 15:     Enable UART2 TXEN (override DIO 10)
 * 0x20: DIO group 1 DIO[14:0] input data (bits 14-0)
 * 0x22: DIO group 2 DIO[26:22] DIO[20:15] input data (bits 10-0)
 * 0x24: DIO group 3 DIO[42:27] input data (bits 15-0)
 * 0x26: DIO group 4 DIO[59:48] input data (bits 11-0)
 * 0x28: TAG memory access
 * 0x2a: Custom load ID register (reads 0 on standard load)
 * 0x2c: IRQ register (smc_irq line indicates one of the following is active)
 *   bit 0: XUART IRQ
 *   bit 1: CAN IRQ
 *   bit 2: CAN2 IRQ
 *   bit 3: PC/104 IRQ 5
 *   bit 4: PC/104 IRQ 6
 *   bit 5: PC/104 IRQ 7
 *   bits 6-15: reserved
 * 0x2e: IRQ mask register.  Setting a bit in this register disables the IRQ
 *       for the corresponding bit in the IRQ register.
 */
 
/*
 *  0 - Initial revision
 *  1 - Move DIO reset switch function to the correct DIO
 *  2 - Enable IRQs
 */

module syscon(
  wb_clk_i,
  wb_rst_i,

  wb_cyc_i,
  wb_stb_i,
  wb_we_i,
  wb_adr_i,
  wb_dat_i,
  wb_sel_i,
  wb_dat_o,
  wb_ack_o,

  dio_i,
  dio_oe_o,
  dio_o,
  usec_i,
  mode1_i,
  mode2_i,

  bus_config_o,
  en_can_o,
  en_can2_o,
  en_spi_o,
  en_bbclk_o,
  en_txen_o,
  en_touch_o,
  red_led_o,
  green_led_o,
  reboot_o,
  irqs_i,
  evgpio_irq_o,
  irq_o,
  sd_busy_i,
  dma_idle_i

);

input wb_clk_i, wb_rst_i, wb_cyc_i, wb_stb_i, wb_we_i;
input [31:0] wb_adr_i;
input [15:0] wb_dat_i;
input [1:0] wb_sel_i;
output [15:0] wb_dat_o;
output wb_ack_o;

input [59:0] dio_i;
output [59:0] dio_oe_o;
output [59:0] dio_o;

input [31:0] usec_i;
input mode1_i;
input mode2_i;

output [15:0] bus_config_o;
output en_can_o;
output en_can2_o;
output en_spi_o;
output en_bbclk_o;
output [5:0] en_txen_o;
output en_touch_o;
output red_led_o;
output green_led_o;
output reboot_o;
output evgpio_irq_o;
input [15:0] irqs_i;
output irq_o;
input dma_idle_i, sd_busy_i;

localparam [15:0] model = 16'h7700;
localparam [3:0] revision = 4'h2;
localparam [15:0] custom = 16'h0000;

reg tag_csn, tag_clk, tag_si;
wire tag_so;
reg tag_so_q;
tagmem tagmemcore(
  .CLK(tag_clk),
  .SI(tag_si),
  .SO(tag_so),
  .CS(tag_csn)
);

always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) tag_so_q <= 1'b0;
  else tag_so_q <= tag_so;
end

wire random_clk;
OSCE #(.NOM_FREQ("3.1")) rndosc (random_clk);
reg [31:0] random;
reg [15:0] rndcnt;
reg [1:0] wdogctl;
reg [9:0] wdogcnt;
reg feed_en, reboot;
wire sed_err;
wire feed_bsy, feed_rdreq;
wire [15:0] wdog_dat;
reg resetsw_en;
assign reboot_o = reboot | (resetsw_en & !dio_i[29]);
always @(posedge random_clk or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    rndcnt <= 16'd0;
    random <= 32'd0;
    wdogctl <= 2'd1;   
    wdogcnt <= 10'd0;
  end else begin
    rndcnt <= rndcnt + 1'b1;
    if (rndcnt == 16'd0) begin
      random[31:0] <= {random[30:0], wb_clk_i};
      wdogcnt <= wdogcnt + 1'b1;
    end
    if (feed_rdreq) begin
      wdogctl <= wdog_dat;
      wdogcnt <= 10'd0;
    end
  end
end
always @(*) begin
  case (wdogctl)
  2'd0: reboot = wdogcnt[4]; /* approx .338 seconds */
  2'd1: reboot = wdogcnt[7]; /* approx 2.706 seconds */
  2'd2: reboot = wdogcnt[9]; /* approx 10.824 seconds */
  2'd3: reboot = 1'b0;
  endcase
end

resync_fifo watchdogfifo(
  .rst_i(wb_rst_i),
  .wrclk_i(wb_clk_i),
  .wrreq_i(feed_en),
  .data_i(wb_dat_i),
  .full_wrclk_o(feed_bsy),
  .rdclk_i(random_clk),
  .rdreq_i(feed_rdreq),
  .q_o(wdog_dat),
  .full_rdclk_o(feed_rdreq)
);

reg [1:0] scratch;

reg [15:0] bus_config;
reg en_can, en_can2, en_spi, en_bbclk, en_touch;
reg [5:0] en_txen; 
reg red_led, green_led;
reg [15:0] irq_mask;

assign bus_config_o = bus_config;
assign en_can_o = en_can;
assign en_can2_o = en_can2;
assign en_spi_o = en_spi;
assign en_bbclk_o = en_bbclk;
assign en_txen_o = en_txen;
assign en_touch_o = en_touch;
assign red_led_o = red_led;
assign green_led_o = green_led;

assign irq_o = ((irqs_i & ~irq_mask) != 16'h0000);

reg [15:0] tagmem0, tagmem1;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    en_can <= 1'b0;
    en_can2 <= 1'b0;
    en_spi <= 1'b0;
    tag_clk <= 1'b0;
    tag_si <= 1'b0;
    tag_csn <= 1'b0;
    en_bbclk <= 1'b0;
    en_txen <= 6'd0;
    en_touch <= 1'b0;
    bus_config <= 16'h0000;
    scratch <= 2'b00;
    red_led <= 1'b1;
    green_led <= 1'b1;
    resetsw_en <= 1'b0;
    irq_mask <= 16'h0000;
    tagmem0 <= 16'h0;
    tagmem1 <= 16'h0;
  end else begin
    if (wb_cyc_i && wb_stb_i && wb_we_i) case (wb_adr_i[5:0])
      6'h2: {resetsw_en, en_touch, en_txen[4], en_txen[0], en_bbclk, en_spi,
             en_can2, en_can, scratch} <= wb_dat_i[15:6];
      6'h4: bus_config <= wb_dat_i;
      6'h12: {red_led, green_led} <= wb_dat_i[12:11];
      6'h16: {en_txen[2:1], en_txen[5], en_txen[3]} <= wb_dat_i[15:12];
      6'h28: {tag_clk, tag_si, tag_csn} <= wb_dat_i[3:1];
      6'h2e: irq_mask <= wb_dat_i;
      6'h30: tagmem0 <= wb_dat_i;
      6'h32: tagmem1 <= wb_dat_i;
    endcase
  end
end

reg [15:0] wb_dat;
assign wb_dat_o = wb_dat;
reg wb_ack, evgpio_en, evgpiomask_en;
wire [15:0] evgpio_dat, evgpiomask_dat;
wire evgpio_ack, evgpio_irq, evgpiomask_ack;
wire [127:0] evgpio_o, evgpiomask_o;
assign wb_ack_o = wb_ack;
assign dio_o[59:0] = evgpio_o[59:0];
assign dio_oe_o[59:0] = evgpio_o[123:64];

wb_evgpio evgpiocore(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wb_cyc_i && evgpio_en),
  .wb_stb_i(wb_stb_i && evgpio_en),
  .wb_we_i(wb_we_i),
  .wb_dat_o(evgpio_dat),
  .wb_dat_i(wb_dat_i),
  .wb_ack_o(evgpio_ack),

  .gpios_i(dio_i & (~evgpiomask_o)),
  .gpios_o(evgpio_o),
  .irq_o(evgpio_irq_o)
);

wb_evgpio evgpiomaskcore(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wb_cyc_i && evgpiomask_en),
  .wb_stb_i(wb_stb_i && evgpiomask_en),
  .wb_we_i(wb_we_i),
  .wb_dat_o(evgpiomask_dat),
  .wb_dat_i(wb_dat_i),
  .wb_ack_o(evgpiomask_ack),

  .gpios_o(evgpiomask_o)
);

always @(*) begin
  wb_ack = wb_cyc_i && wb_stb_i;
  wb_dat = 16'hxxxx;
  evgpio_en = 1'b0;
  evgpiomask_en = 1'b0;

  case (wb_adr_i[5:0]) 
    6'h0: wb_dat = model[15:0];
    6'h2: wb_dat = {resetsw_en, en_touch, en_txen[4], en_txen[0], en_bbclk,
          en_spi, en_can2, en_can, scratch, mode2_i, mode1_i, revision};
    6'h4: wb_dat = bus_config;
    6'h6: begin
      feed_en = wb_cyc_i & wb_stb_i & wb_we_i & !feed_bsy;
      wb_ack = wb_cyc_i & wb_stb_i & ((wb_we_i & !feed_bsy ) | !wb_we_i);
    end
    6'h8: wb_dat = usec_i[15:0];
    6'ha: wb_dat = usec_i[31:16];
    6'hc: wb_dat = random[15:0];
    6'he: wb_dat = random[31:16];
    6'h10: begin
      wb_dat = evgpio_dat;
      evgpio_en = 1'b1;
      wb_ack = evgpio_ack;
    end
    6'h12: wb_dat = {sd_busy_i, dma_idle_i, 1'bx, red_led, green_led, {11{1'bx}}};
    6'h14: begin
      wb_dat = evgpiomask_dat;
      evgpiomask_en = 1'b1;
      wb_ack = evgpiomask_ack;
    end
    6'h16: wb_dat = {en_txen[2:1], en_txen[5], en_txen[3], {12{1'bx}}};
    6'h20: wb_dat = {1'bx, dio_i[14:0]};
    6'h22: wb_dat = {5'bxxxxx, dio_i[26:22], dio_i[20:15]};
    6'h24: wb_dat = dio_i[42:27];
    6'h26: wb_dat = {4'bxxxx, dio_i[59:48]};
    6'h28: wb_dat = {8'hxx, 2'b00, wdogctl, tag_clk, tag_si, tag_csn, tag_so_q};
    6'h2a: wb_dat = custom;
    6'h2c: wb_dat = irqs_i & ~irq_mask;
    6'h2e: wb_dat = irq_mask;
    6'h30: wb_dat = tagmem0;
    6'h32: wb_dat = tagmem1;
    //6'h2c: wb_dat = wb_sel_i[1:0] == 2'b11 ? 16'h1616 : 16'h8888; //width test
    //default: wb_dat = 16'hxxxx;
  endcase
end

endmodule

